float Script : STANDARDSGLOBAL <
    string   UiControl = "none";
    string   ScriptClass = "scene";
    string   ScriptOrder = "standard";
    string   ScriptOutput = "colour";
    string   Script = "Technique=PBR_Metal_Rough_Spherical;";
    string   ScriptDescription = "<b>PBR that uses metalness and roughness maps with spherical environment maps</b>";
> = 0.1; // version #
 
int nbSamples
<
    string   UiName = "nbSamples";
    string   UiControl = "Slider";
   float UiMin = 4.0f;
   float UiMax = 64.0f;
 > = 16;

float envExposure
<
    string   UiName = "Env Exposure";
    string   UiControl = "Slider";
   float UiMin = -10.0f;
   float UiMax = 10.0f;
 > = 1;
 
float envRotation
<
    string   UiName = "Env Rotation";
    string   UiControl = "Slider";
   float UiMin = -99999999999.0f;
   float UiMax = 99999999999.0f;
 > = 0;

sampler2D tex0 <  string BindAddress = "Tog.Texture[0].2D"; >;			///albedo
sampler2D tex1 <  string BindAddress = "Tog.Texture[1].2D"; >;		///Environment
sampler2D tex2 <  string BindAddress = "Tog.Texture[2].2D"; >;			///metal/rough/AO (rgb)
sampler2D tex3 <  string BindAddress = "Tog.Texture[3].2D"; >;			///Normal
float4x4 worldMatrix <	 string BindAddress = "Tog.Camera.Model"; >;
float4x4 invTransWorldMatrix <  string BindAddress = "Tog.Camera.InverseTransposeModel";>;
float4x4 inverseViewMatrix <  string BindAddress = "Tog.Camera.InverseView";>;

float4 cameraPositionWS <  string BindAddress = "Tog.Camera.Location";>;

uniform float4  clip04Enabled       < string BindAddress = "Tog.ClipPlane04Enabled"; >;
uniform float4  clip04SoftDistance  < string BindAddress = "Tog.ClipPlane04SoftDistance"; >;

const float M_PI = 3.1415926535897932384626433832795;
const float M_INV_PI = 0.31830988618379067153776752674503;
const float M_INV_LOG2 = 1.4426950408889634073599246810019;
const int maxLod = 11;

float dynamicSpecularFactorC = 5;
float specularReflectionRange = 0.6f;
float DSFExposureScale = 1.1;
float exponent = 6.0f;
float DSFESwitch = 1.0;

float radicalInverse[64] = {0, 0.5, 0.25, 0.75, 0.125, 0.625, 0.375, 0.875, 0.0625, 0.5625, 0.3125, 0.8125, 0.1875, 0.6875, 0.4375, 0.9375, 0.03125, 0.53125, 0.28125, 0.78125, 0.15625, 0.65625, 0.40625, 0.90625, 0.09375, 0.59375, 0.34375, 0.84375, 0.21875, 0.71875, 0.46875, 0.96875, 0.015625, 0.515625, 0.265625, 0.765625, 0.140625, 0.640625, 0.390625, 0.890625, 0.078125, 0.578125, 0.328125, 0.828125, 0.203125, 0.703125, 0.453125, 0.953125, 0.046875, 0.546875, 0.296875, 0.796875, 0.171875, 0.671875, 0.421875, 0.921875, 0.109375, 0.609375, 0.359375, 0.859375, 0.234375, 0.734375, 0.484375, 0.984375};

struct App2vert
{
	float3 position : POSITION;
	float3 normal : NORMAL;
	float3 tangent : TANGENT;
	float3 bitangent : BINORMAL;	
	float4 texcoord0 : TEXCOORD0;
	float4 texcoord1 : TEXCOORD1;
	float4 texcoord2 : TEXCOORD2;
	float4 texcoord3 : TEXCOORD3;

};

struct Vert2frag
{
	float4 hpos : POSITION;
	float4 tex_coord : TEXCOORD0;
	float4 positionWS : TEXCOORD1;
	float3 normalWS : TEXCOORD2;
	float3 tangentWS : TEXCOORD3;
	float3 bitangentWS : TEXCOORD4;
	float4 clipValues : TEXCOORD5;
};

Vert2frag vertex_main(
	App2vert vs_in)
{
	Vert2frag vs_out;
	float4 position = float4(vs_in.position, 1);
	//WVP * position
	vs_out.hpos = mul(glstate.matrix.mvp, position);
	//position
	vs_out.positionWS = mul(worldMatrix, position);
	//normal, tangent, bitangent

	vs_out.normalWS = normalize(mul((float3x3)invTransWorldMatrix, vs_in.normal));
	vs_out.tangentWS = normalize(mul((float3x3)invTransWorldMatrix, vs_in.tangent));
	vs_out.bitangentWS = normalize(mul((float3x3)invTransWorldMatrix, vs_in.bitangent));
	//texcoord
	vs_out.tex_coord = vs_in.texcoord0;
	
	vs_out.clipValues = 0.0;

	float4 eyePosVS = mul(glstate.matrix.modelview[0], position);

	if(clip04Enabled.x == 1.0f)
		vs_out.clipValues.x = dot(eyePosVS.xyz,glstate.clip[0].plane.xyz) + glstate.clip[0].plane.w;
	if(clip04Enabled.y == 1.0f)
		vs_out.clipValues.y = dot(eyePosVS.xyz,glstate.clip[1].plane.xyz) + glstate.clip[1].plane.w;
	if(clip04Enabled.z == 1.0f)
		vs_out.clipValues.z = dot(eyePosVS.xyz,glstate.clip[2].plane.xyz) + glstate.clip[2].plane.w;
	if(clip04Enabled.w == 1.0f)
		vs_out.clipValues.w = dot(eyePosVS.xyz,glstate.clip[3].plane.xyz) + glstate.clip[3].plane.w;
    
	return vs_out;
}

float normal_distrib(
	float ndh,
	float Roughness){
  // use GGX / Trowbridge-Reitz, same as Disney and Unreal 4
  // cf http://blog.selfshadow.com/publications/s2013-shading-course/karis/s2013_pbs_epic_notes_v2.pdf p3
  float alpha = Roughness * Roughness;
  float tmp = alpha / max(1e-8,(ndh*ndh*(alpha*alpha-1.0)+1.0));
  return tmp * tmp * M_INV_PI;
}

float3 normalBlend(
	float3 normA,
	float3 normB){
	return normalize(float3(normA.x + normB.x, normA.y + normB.y, normA.z * normB.z));
}

float probabilityGGX(
	float ndh,
	float vdh,
	float Roughness){
  return normal_distrib(ndh, Roughness) * ndh / (4.0*vdh);
}

float distortion(
	float3 Wn){
  // Computes the inverse of the solid angle of the (differential) pixel in
  // the cube map pointed at by Wn
  float sinT = sqrt(1.0-Wn.y*Wn.y);
  return sinT;
}

float computeLOD(
	float3 Ln,
	float p){
  return max(0.0, (maxLod-1.5) - 0.5*(log(float(nbSamples)) + log( p * distortion(Ln) ))	* M_INV_LOG2);
}

float3 envSampleLOD(
	float3 dirWS,
	float lod){
	
	//spherical mapping
	float n = length(dirWS.xz);
	float4 pos = float4( (n>0.0000001) ? dirWS.x / n : 0.0, dirWS.y, 0, 0);
	pos = acos(pos)*M_INV_PI;
	pos.x = (dirWS.z > 0.0) ? pos.x*0.5 : 1.0-(pos.x*0.5);
	pos.x += envRotation/360;
	pos.w = lod;
	pos.y = 1.0-pos.y;
	float3 hdrSample = tex2Dlod(tex1, pos).rgb * pow(2, envExposure);
	return hdrSample;

	//cube mapping
	/*float3 rotatedUVW = dirWS;
	rotatedUVW.x = (cos(radians(envRotation)) * dir.x) + (sin(radians(envRotation)) * dir.z);
	rotatedUVW.y = dir.y;
	rotatedUVW.z = (-1 * (sin(radians(envRotation)) * dir.x)) + (cos(radians(envRotation)) * dir.z);
	return texCUBElod(tex1, float4(rotatedUVW.xyz, max(1, lod))).rgb * envExposure;*/
}

float3 envIrradiance(float3 dirWS)
{
	// JASON HACK
	//
	// Just base irradiance on the exposure...
	return float3(1.0) * pow(2, envExposure);

 /* float rot = environment_rotation * M_2PI;
  float crot = cos(rot);
  float srot = sin(rot);
  float4 shDir = float4(dirWS.xzy, 1.0);
  shDir = float4(
	shDir.x * crot - shDir.y * srot,
	shDir.x * srot + shDir.y * crot,
	shDir.z,
	1.0);
  return float3(
	dot(shDir, irrad_mat_red * shDir),
	dot(shDir, irrad_mat_green * shDir),
	dot(shDir, irrad_mat_blue * shDir)
	) * environment_exposure;*/
}

float G1(
	float ndw, // w is either Ln or Vn
	float k){
  // One generic factor of the geometry function divided by ndw
  // NB : We should have k > 0
  return 1.0 / ( ndw*(1.0-k) +  k );
}

float3 fresnel(
	float vdh,
	float3 F0)
{
  // Schlick with Spherical Gaussian approximation
  // cf http://blog.selfshadow.com/publications/s2013-shading-course/karis/s2013_pbs_epic_notes_v2.pdf p3
  float sphg = pow(2.0, (-5.55473*vdh - 6.98316) * vdh);
  return F0 + (vec3(1.0, 1.0, 1.0) - F0) * sphg;
}

float visibility(
	float ndl,
	float ndv,
	float Roughness){
  // Schlick with Smith-like choice of k
  // cf http://blog.selfshadow.com/publications/s2013-shading-course/karis/s2013_pbs_epic_notes_v2.pdf p3
  // visibility is a Cook-Torrance geometry function divided by (n.l)*(n.v)
  float k = max(Roughness * Roughness * 0.5, 1e-5);
  return G1(ndl,k)*G1(ndv,k);
}

float3 cook_torrance_contrib(
	float vdh,
	float ndh,
	float ndl,
	float ndv,
	float3 Ks,
	float Roughness){
  // This is the contribution when using importance sampling with the GGX based
  // sample distribution. This means ct_contrib = ct_brdf / ggx_probability
  return fresnel(vdh,Ks) * (visibility(ndl,ndv,Roughness) * vdh * ndl / ndh );
}

float3 importanceSampleGGX(
	float2 Xi,
	float3 A,
	float3 B,
	float3 C,
	float roughness)
{
	float a = roughness*roughness;
	float cosT = sqrt((1.0-Xi.y)/(1.0+(a*a-1.0)*Xi.y));
	float sinT = sqrt(1.0-cosT*cosT);
	float phi = 2.0*M_PI*Xi.x;
	return (sinT*cos(phi)) * A + (sinT*sin(phi)) * B + cosT * C;
}

float radicalInverse_VdC(
uint bits){
     bits = (bits << 16u) | (bits >> 16u);
     bits = ((bits & 0x55555555u) << 1u) | ((bits & 0xAAAAAAAAu) >> 1u);
     bits = ((bits & 0x33333333u) << 2u) | ((bits & 0xCCCCCCCCu) >> 2u);
     bits = ((bits & 0x0F0F0F0Fu) << 4u) | ((bits & 0xF0F0F0F0u) >> 4u);
     bits = ((bits & 0x00FF00FFu) << 8u) | ((bits & 0xFF00FF00u) >> 8u);
     return float(bits) * 2.3283064365386963e-10; // / 0x100000000
}

float2 hammersley2D(
	int i,
	int nbSamples){
	return float2((float)i / (float)nbSamples, radicalInverse[i]);
}

float4 PBR_Main(
	Vert2frag in_state,
	sampler2D in_albedoTexture,
	sampler2D in_environmentMap,
	sampler2D in_normalTexture,
	sampler2D in_AO_Metal_RoughTexture){
	
	float ao, roughness, metallic;
	float3 normalMap, baseColour;
	float4 out_colour;
	
	  //Clipping + Soft Clipping
  float clipAlpha = 1.0f;
  if(any(clip04Enabled))
  {
    if(any(clip04SoftDistance))
    {
      in_state.clipValues += max(-clip04SoftDistance,0);
      float4 clips;
      clips = in_state.clipValues*(1.0/abs(clip04SoftDistance));
      clipAlpha = clamp(min(min(min(clips.x,clips.y),clips.z),clips.w),0,1);
    }
    clip(in_state.clipValues);
  }
  
	//float4 of (2.2, 2.2, 2.2, 1.0) used to transform sampled texels from Gamma>Linear space
	float4 gammaFactor = float4(2.2, 2.2, 2.2, 1.0);//2.2
		
	//gets Gamma space Albedo colour for this texel, Gamma>Linear space transform, scale by Albedo Multiplier
	baseColour = pow(tex2D(in_albedoTexture, in_state.tex_coord.xy), gammaFactor).rgb;//rgba to get alpha?
	float baseAlpha = tex2D(in_albedoTexture, in_state.tex_coord.xy).a;
	
	//gets normal map value for this texel, transforms it to tangent space
	float3 normalTS = tex2D(in_normalTexture, in_state.tex_coord.xy).xyz;
	normalTS = (normalTS.xyz - 0.5) * 2.0;

	// Get ao, rought and metal from texture
	float3 ao_metal_rough = tex2D(in_AO_Metal_RoughTexture, in_state.tex_coord.xy);
	ao = ao_metal_rough.r;
	metallic = ao_metal_rough.g;
	roughness = ao_metal_rough.b;

	float3 eyeVecWS = normalize(cameraPositionWS - in_state.positionWS).xyz;
	
	float3 normalMapWS = normalize(
	normalTS.x * in_state.tangentWS
	 + normalTS.y * in_state.bitangentWS
	 + normalTS.z * in_state.normalWS
	);
	
	float3 dielectricColour = float3(0.04);
	
	float ao_intensity = 1.0;
	ao = lerp(1.0, ao, ao_intensity * (1.0 - metallic));

	//balances the base colour between diff and spec based on metalness
	float3 diffColour = baseColour * (1.0 - metallic);
	float3 specColour = lerp(dielectricColour, baseColour, metallic);
	  
    float3 TpWS = normalize(in_state.tangentWS - (normalMapWS * dot(in_state.tangentWS, normalMapWS))); // local tangent
	float3 BpWS = normalize(in_state.bitangentWS - 
							(normalMapWS * dot(in_state.bitangentWS, normalMapWS)) - 
							(TpWS * dot(in_state.bitangentWS, TpWS))
						); // local bitangent


	float ndv = dot(eyeVecWS, normalMapWS);

	if (ndv < 0) 
	{
		eyeVecWS = reflect(eyeVecWS, normalMapWS);
		ndv = abs(ndv);
  	}

  float3 contribE = envIrradiance(normalMapWS) * diffColour * ao;

  float3 contribS = float3(0.0);
  
  for(int i=0; i<nbSamples; ++i)
  {
	float2 Xi = hammersley2D(i, nbSamples);
	float3 HnWS = importanceSampleGGX(Xi,TpWS,BpWS,normalMapWS,roughness);
	float3 LnWS = -reflect(eyeVecWS, HnWS);
	
	float ndl = dot(normalMapWS, LnWS);

	// Horizon fading trick from http://marmosetco.tumblr.com/post/81245981087
	const float horizonFade = 1.3;
	float horiz = clamp( 1.0 + horizonFade * ndl, 0.0, 1.0 );
	horiz *= horiz;
	
	ndl = max( 1e-8, abs(ndl) );

	float vdh = max(1e-8, dot(eyeVecWS, HnWS));
	float ndh = max(1e-8, dot(normalMapWS, HnWS));

	float lodS = roughness < 0.01 ? 0.0 : computeLOD(LnWS, probabilityGGX(ndh, vdh, roughness));

	contribS += envSampleLOD(LnWS, lodS) * cook_torrance_contrib(vdh, ndh, ndl, ndv, specColour, roughness) * horiz * ao;
  }
  contribS /= float(nbSamples);

  // Emissive
  float3 contribEm =  0/*emissive_intensity * texture2D(emissive_tex, in_state.tex_coord).rgb*/;//not yet used

float3 specularBoost;
float avg = (contribE.r + contribE.g + contribE.b)/3;
specularBoost.r = avg * contribS.r;
specularBoost.g = avg * contribS.g;
specularBoost.b = avg * contribS.b;
specularBoost *= (dynamicSpecularFactorC * (DSFESwitch * pow(DSFExposureScale, envExposure))) + ((1 - DSFESwitch) * dynamicSpecularFactorC);
  
  
	
	float combinedE = (contribE.r * 0.2989) + (contribE.g * 0.587) + (contribE.b * 0.114);

  
  // diff atten
  contribE.r *= baseAlpha;
  contribE.g *= baseAlpha;
  contribE.b *= baseAlpha;
  
  //spec atten
  contribS += specularBoost;
  contribS *= ((1.0f - specularReflectionRange) + (baseAlpha * specularReflectionRange));
  float combinedS = (contribS.r * 0.2989) + (contribS.g * 0.587) + (contribS.b * 0.114);
  //outalpha calc
	float outAlpha = (combinedS / (combinedE + combinedS + 0.5));
	float specularAlphaContribution = pow(combinedS, exponent) ;
	outAlpha = baseAlpha + ((1.0f - baseAlpha) * specularAlphaContribution);
	
	//old method, provided better visual results but was somewhat broken in certain cases.
	//outAlpha = baseAlpha + ( ((1.0f - baseAlpha) * (1.0f * (combinedS / (combinedE + combinedS)))));

	float4 final = float4(contribS + contribE + contribEm, outAlpha);
	return final;
}

float4 fragment_main(
	Vert2frag fs_in) : COLOR
{	
	//float4 fs_out = float4(pow(PBR_Main(fs_in, tex0, tex1, tex3, tex2), (1 / 2.2)), glstate.material.front.diffuse.a);
	float4 fs_out = PBR_Main(fs_in, tex0, tex1, tex3, tex2);
	fs_out.r = pow(fs_out.r, (1/2.2));
	fs_out.g = pow(fs_out.g, (1/2.2));
	fs_out.b = pow(fs_out.b, (1/2.2));
	fs_out.a = fs_out.a * glstate.material.front.diffuse.a;
	return fs_out;
}

technique PBR_Metal_Rough_Spherical
{
	pass P0
	{
		DepthTestEnable = true;
		DepthMask       = true;
		CullFaceEnable  = true;
		BlendEnable = true;
		blendfunc = int2(srcalpha, oneminussrcalpha);
		VertexShader = compile vp40 vertex_main();
		PixelShader  = compile gp4fp fragment_main(); // gp4fp
	}
}

 string description = "PBRTestConversion";